""" "Module for authenticating with the Zabbix API as well as
loading/storing authentication information.

Manages the following:
- Loading and saving auth token files (file containing API session token)
- Loading and saving auth files (file containing username and password)
- Loading username and password from environment variables
- Prompting for username and password
- Logging in to the Zabbix API using one of the above methods
"""

from __future__ import annotations

import logging
import os
import sys
from collections.abc import Generator
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Final
from typing import NamedTuple
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import PrivateAttr
from pydantic import RootModel
from rich.console import ScreenContext
from strenum import StrEnum

from zabbix_cli._v2_compat import AUTH_FILE as AUTH_FILE_LEGACY
from zabbix_cli._v2_compat import AUTH_TOKEN_FILE as AUTH_TOKEN_FILE_LEGACY
from zabbix_cli.config.constants import AUTH_FILE
from zabbix_cli.config.constants import AUTH_TOKEN_FILE
from zabbix_cli.config.constants import ConfigEnvVars
from zabbix_cli.exceptions import AuthError
from zabbix_cli.exceptions import AuthTokenFileError
from zabbix_cli.exceptions import SessionFileError
from zabbix_cli.exceptions import SessionFileNotFoundError
from zabbix_cli.exceptions import SessionFilePermissionsError
from zabbix_cli.exceptions import ZabbixAPIException
from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.logs import add_user
from zabbix_cli.output.console import err_console
from zabbix_cli.output.console import error
from zabbix_cli.output.console import warning
from zabbix_cli.output.prompts import str_prompt
from zabbix_cli.pyzabbix.client import ZabbixAPI

if TYPE_CHECKING:
    from zabbix_cli.config.model import Config


logger = logging.getLogger(__name__)


SECURE_PERMISSIONS: Final[int] = 0o600
SECURE_PERMISSIONS_STR = format(SECURE_PERMISSIONS, "o")


class SessionInfo(BaseModel):
    """Information about a session for a specific user."""

    username: str
    session_id: str


class SessionList(RootModel[list[SessionInfo]]):
    """List of sessions for a specific Zabbix API URL."""

    root: list[SessionInfo] = []

    def __len__(self) -> int:
        return len(self.root)

    def set_session(self, username: str, session_id: str) -> None:
        """Add or modify a session for a given user."""
        for session in self.root:
            if session.username == username:
                session.session_id = session_id
                return
        # Did not find existing session, add new one
        self.root.append(SessionInfo(username=username, session_id=session_id))

    def get_session(self, username: str) -> Optional[SessionInfo]:
        for session in self.root:
            if session.username == username:
                return session
        return None


class SessionFile(RootModel[dict[str, SessionList]]):
    """Contents of a session file.

    Root model that wraps a dict of URLs to `SessionList` objects.
    """

    root: dict[str, SessionList] = {}
    _path: Optional[Path] = PrivateAttr(default=None)

    def get_sessions(self, url: str) -> SessionList:
        """Get session list for a URL.

        Note:
            Returns a new empty `SessionList` if URL has no sessions.

            The returned `SessionList` should never be mutated directly.
            Always use `set_user_session` to modify a user's session for
            a given URL.
        """
        return self.root.get(url, None) or SessionList()  # lazily create default

    def set_sessions(self, url: str, sessions: SessionList) -> None:
        """Set session list for a URL."""
        self.root[url] = sessions

    def get_user_session(self, url: str, username: str) -> Optional[SessionInfo]:
        """Get a session given a URL and username."""
        return self.get_sessions(url).get_session(username)

    def set_user_session(self, url: str, username: str, session_id: str) -> None:
        """Add or update a user's session for a given URL."""
        session = self.get_sessions(url)
        session.set_session(username, session_id)
        self.set_sessions(url, session)

    @classmethod
    def load(cls, file: Path, *, allow_insecure: bool = False) -> SessionFile:
        """Load the contents of a session file."""
        if not file.exists():
            raise SessionFileNotFoundError("Session file does not exist: %s", file)

        if not file_has_secure_permissions(file) and not allow_insecure:
            raise SessionFilePermissionsError(
                f"Session file {file} must have {SECURE_PERMISSIONS_STR} permissions. Refusing to load."
            )
        try:
            contents = file.read_text()
            cl = cls.model_validate_json(contents)
            cl._path = file
            return cl
        except Exception as e:
            raise SessionFileError(f"Unable to load session file {file}: {e}") from e

    def save(
        self, path: Optional[Path] = None, *, allow_insecure: bool = False
    ) -> None:
        path = path or self._path
        if not path:
            raise SessionFileError("Cannot save session file without a path.")

        try:
            logger.debug("Saving session file %s", path)
            path.write_text(self.model_dump_json(indent=2))
            if not allow_insecure and not file_has_secure_permissions(path):
                set_file_secure_permissions(path)
            logger.info("Saved session file %s", path)
        except Exception as e:
            raise SessionFileError(
                f"Unable to save session file {self._path}: {e}"
            ) from e


class CredentialsType(StrEnum):
    """Login credentials type."""

    PASSWORD = "username and password"
    AUTH_TOKEN = "auth token"
    SESSION = "session"


class CredentialsSource(StrEnum):
    """Source of login credentials."""

    ENV = "env"
    FILE = "file"
    PROMPT = "prompt"
    CONFIG = "config"
    LOGIN_COMMAND = "login command"


class Credentials(NamedTuple):
    """Credentials for logging in to the Zabbix API."""

    source: Optional[CredentialsSource] = None
    username: Optional[str] = None
    password: Optional[str] = None
    auth_token: Optional[str] = None
    session_id: Optional[str] = None

    def __str__(self) -> str:
        if self.type and self.source:
            return f"{self.type} from {self.source}"
        return "no credentials"

    @property
    def type(self) -> Optional[CredentialsType]:
        if self.auth_token:
            return CredentialsType.AUTH_TOKEN
        if self.username and self.password:
            return CredentialsType.PASSWORD
        if self.session_id:
            return CredentialsType.SESSION
        return None

    def is_valid(self) -> bool:
        """Check if credentials are valid (non-empty)."""
        return self.type is not None


class LoginInfo(NamedTuple):
    credentials: Credentials
    token: str
    """Token or Session ID returned from the Zabbix API."""


class Authenticator:
    """Encapsulates logic for authenticating with the Zabbix API
    using various methods, as well as storing and loading auth tokens.


    Exposes methods for logging in with specific credentials, as well as
    a method for logging in with any valid credentials.

    Bootstraps application state with login and API info on successful login.
    """

    config: Config

    def __init__(self, config: Config) -> None:
        self.config = config
        # Ensure we have a Zabbix API URL before instantiating client
        self.config.api.url = self.get_zabbix_url()
        self.client = ZabbixAPI.from_config(self.config)

    @cached_property
    def screen(self) -> ScreenContext:
        return err_console.screen()

    def login_with_any(self) -> tuple[ZabbixAPI, LoginInfo]:
        """Log in to the Zabbix API using any valid method.

        Returns the Zabbix API client object.

        If multiple methods are available, they are tried in the following order:

        1. API token in environment variables
        2. API token in config file
        3. Session ID in session file (if `use_session_file=true`)
        4. Username and password in environment variables
        5. Username and password in config file
        6. Username and password in auth file
        7. Session ID in legacy auth token file (if `use_session_file=true`)
        8. Username and password from prompt
        """
        for credentials in self._iter_all_credentials():
            if not credentials.is_valid():
                logger.debug("No valid credentials found with %s", credentials)
                continue
            info = self.login_with_credentials(credentials)
            if info:
                return self.client, info
        raise AuthError(
            f"No authentication method succeeded for {self.config.api.url}. Check the logs for more information."
        )

    def _iter_all_credentials(
        self, *, prompt_password: bool = True
    ) -> Generator[Credentials, None, None]:
        """Generator of credentials from all possible sources.

        Only yields non-empty credentials, but does not check if they are valid.

        Finally yields a prompt for username and password if `prompt_password=True`.
        """
        for func in [
            self._get_auth_token_env,
            self._get_auth_token_config,
            self._get_session_file,
            self._get_username_password_env,
            self._get_username_password_config,
            self._get_username_password_auth_file,
            self._get_auth_token_file_legacy,
        ]:
            try:
                creds = func()
                if creds:
                    yield creds
            except Exception as e:
                logger.error("Error getting credentials from %s: %s", func, e)
                continue

        if prompt_password:
            yield self._get_username_password_prompt()

    @classmethod
    def login_with_prompt(cls, config: Config) -> tuple[ZabbixAPI, LoginInfo]:
        """Log in to the Zabbix API using username and password from a prompt."""
        auth = cls(config)
        creds = auth._get_username_password_prompt()
        info = auth.login_with_credentials(creds)
        creds._replace(source=CredentialsSource.LOGIN_COMMAND)
        if not info:
            raise AuthError("Failed to log in with username and password.")
        return auth.client, info

    @classmethod
    def login_with_username_password(
        cls, config: Config, username: str, password: str
    ) -> tuple[ZabbixAPI, LoginInfo]:
        """Log in to the Zabbix API using username and password from a prompt."""
        auth = cls(config)
        creds = Credentials(
            username=username,
            password=password,
            source=CredentialsSource.LOGIN_COMMAND,
        )
        info = auth.login_with_credentials(creds)
        if not info:
            raise AuthError("Failed to log in with username and password.")
        return auth.client, info

    @classmethod
    def login_with_token(
        cls, config: Config, token: str
    ) -> tuple[ZabbixAPI, LoginInfo]:
        """Log in to the Zabbix API using username and password from a prompt."""
        auth = cls(config)
        creds = Credentials(
            auth_token=token,
            source=CredentialsSource.LOGIN_COMMAND,
        )
        info = auth.login_with_credentials(creds)
        if not info:
            raise AuthError("Failed to log in with auth token.")
        return auth.client, info

    def login_with_credentials(self, credentials: Credentials) -> LoginInfo | None:
        """Log in to the Zabbix API using the provided credentials.

        Cannot fail; logs errors and returns None if unsuccessful.

        Args:
            credentials (Credentials): Credentials to use for logging in.

        Returns:
            LoginInfo | None: Login information if successful, None if not.
        """
        try:
            logger.debug(
                "Attempting to log in with %s from %s",
                credentials.type,
                credentials.source,
            )
            return self._do_login_with_credentials(credentials)
        except ZabbixAPIException as e:
            logger.warning("Failed to log in with %s: %s", credentials, e)
            return
        except Exception as e:
            logger.error("Unexpected error logging in with %s: %s", credentials, e)
            return

    def _do_login_with_credentials(self, credentials: Credentials) -> LoginInfo:
        """Login to Zabbix API, and update application state if successful."""
        token = self.client.login(
            user=credentials.username,
            password=credentials.password,
            auth_token=credentials.auth_token,
            session_id=credentials.session_id,
        )
        info = LoginInfo(credentials, token)

        if info.credentials.type == CredentialsType.AUTH_TOKEN:
            logger.info("Logged in using auth token from %s", info.credentials.source)
        elif info.credentials.type == CredentialsType.SESSION:
            logger.info("Logged in using session ID from %s", info.credentials.source)
        else:
            logger.info(
                "Logged in as %s using username and password from %s",
                info.credentials.username,
                info.credentials.source,
            )

        self._update_application_with_login_info(info)
        return info

    def _update_application_with_login_info(self, info: LoginInfo) -> None:
        """Update the application state with the login information.

        Includes the following:
        - Write session file if configured
        - Add username to log records context
        - Update config with credentials
        - Set Zabbix API version on the TableRenderable base class
        """
        from zabbix_cli.models import TableRenderable

        # Write session file
        if info.credentials.username and self.config.app.use_session_file:
            sessionfile = self.load_session_file()
            if not sessionfile:
                logger.debug("No session file found, creating new one.")
                sessionfile = SessionFile()

            sessionfile.set_user_session(
                url=self.config.api.url,
                username=info.credentials.username,
                session_id=info.token,
            )
            try:
                sessionfile.save(
                    self.config.app.session_file,
                    allow_insecure=self.config.app.allow_insecure_auth_file,
                )
            except SessionFileError as e:
                logger.error("Unable to save session file: %s", e)

        # Log context
        if info.credentials.username:
            add_user(info.credentials.username)
        else:
            logger.debug("No username detected, adding <TOKEN> to logs")
            add_user("<TOKEN>")

        # Update config with the new credentials
        self._update_config(info.credentials)

        # Set the Zabbix API version on the TableRenderable base class
        TableRenderable.zabbix_version = self.client.version

    def _update_config(self, credentials: Credentials) -> None:
        """Update the config with credentials from the successful login."""
        from pydantic import SecretStr

        if credentials.username:
            self.config.api.username = credentials.username

        # Only update secrets if they were already set in config and new ones are different.
        # we do not want to assign secrets to a config that does not already have them.
        # I.e. user logs in via file, prompt, env var, etc., which should not
        # assign those secrets to the config. ONLY prompt should assign secrets.
        # TODO: Better introspection of login method to determine if secrets should be updated.
        if (
            # we have a token in the config file
            (config_password := self.config.api.password.get_secret_value())
            and credentials.password
            and config_password != credentials.password
        ):
            self.config.api.password = SecretStr(credentials.password)

        if (
            # we have a token in the config file
            (config_token := self.config.api.auth_token.get_secret_value())
            and credentials.auth_token
            and config_token != credentials.auth_token
        ):
            self.config.api.auth_token = SecretStr(credentials.auth_token)

    def get_zabbix_url(self, *, prompt: bool = True) -> str:
        """Get the URL of the Zabbix server from env, config, then finally prompt for it.."""
        for source in [self._get_zabbix_url_env, self._get_zabbix_url_config]:
            url = source()
            if url:
                return url
        if prompt:
            with self.screen:
                return self._get_zabbix_url_prompt()
        raise AuthError("No Zabbix URL found in environment or config.")

    def _get_zabbix_url_env(self) -> str:
        """Get the URL of the Zabbix server from the environment."""
        return os.environ.get(ConfigEnvVars.URL, "")

    def _get_zabbix_url_config(self) -> str:
        """Get the URL of the Zabbix server from the config file."""
        return self.config.api.url

    def _get_zabbix_url_prompt(self) -> str:
        """Prompt for the URL of the Zabbix server."""
        return str_prompt("Zabbix URL (without /api_jsonrpc.php)", empty_ok=False)

    def _get_username_password_env(self) -> Credentials:
        """Get username and password from environment variables."""
        return Credentials(
            username=os.environ.get(ConfigEnvVars.USERNAME),
            password=os.environ.get(ConfigEnvVars.PASSWORD),
            source=CredentialsSource.ENV,
        )

    def _get_auth_token_env(self) -> Credentials:
        """Get auth token from environment variables."""
        return Credentials(
            auth_token=os.environ.get(ConfigEnvVars.API_TOKEN),
            source=CredentialsSource.ENV,
        )

    def _get_username_password_auth_file(
        self,
    ) -> Credentials:
        """Get username and password from environment variables."""
        path, contents = self.load_auth_file()
        if path:
            logger.debug("Loaded auth file %s", path)
        username, password = _parse_auth_file_contents(contents)
        return Credentials(
            username=username,
            password=password,
            source=CredentialsSource.FILE,
        )

    def _get_username_password_config(
        self,
    ) -> Credentials:
        """Get username and password from config file."""
        return Credentials(
            username=self.config.api.username,
            password=self.config.api.password.get_secret_value(),
            source=CredentialsSource.CONFIG,
        )

    def _get_username_password_prompt(
        self,
    ) -> Credentials:
        """Get username and password from a prompt in a separate screen."""
        with self.screen:
            username = str_prompt(
                "Username", default=self.config.api.username, empty_ok=False
            )
            password = str_prompt("Password", password=True, empty_ok=False)
            return Credentials(
                username=username, password=password, source=CredentialsSource.PROMPT
            )

    def _get_auth_token_config(self) -> Credentials:
        """Get auth token from the config file."""
        return Credentials(
            auth_token=self.config.api.auth_token.get_secret_value(),
            source=CredentialsSource.CONFIG,
        )

    def _get_session_file(self) -> Optional[Credentials]:
        """Get session ID from a file."""
        if not self.config.app.use_session_file:
            logger.debug("Not configured to use auth token file.")
            return

        sessionfile = self.load_session_file()
        if not sessionfile:
            return
        logger.info("Found session file %s", sessionfile._path)  # pyright: ignore[reportPrivateUsage]

        session = sessionfile.get_user_session(
            self.config.api.url, self.config.api.username
        )
        if session:
            return Credentials(
                username=session.username,
                session_id=session.session_id,
                source=CredentialsSource.FILE,
            )

    def _get_auth_token_file_legacy(self) -> Optional[Credentials]:
        """Get auth token (session ID) from a legacy auth token file.

        From Zabbix-CLI 3.5.0 onwards, we use a new session file.
        """
        if not self.config.app.use_session_file:
            logger.debug("Not configured to use auth token file.")
            return

        path, contents = self.load_auth_token_file()
        username, auth_token = _parse_auth_file_contents(contents)

        # Found token, but does not match configured username
        if auth_token and username and username != self.config.api.username:
            warning(
                f"Ignoring existing auth token in auth file {path}: "
                f"Username {username!r} in file does not match username {self.config.api.username!r} in configuration file."
            )
            return

        # NOTE: Originally, this was named auth token file, but it does in fact
        # contain a Session ID, not an API token.
        # So we return it under the `session_id` key.
        return Credentials(
            username=username, session_id=auth_token, source=CredentialsSource.FILE
        )

    def load_session_file(self) -> Optional[SessionFile]:
        """Load a session file from configured path."""
        try:
            return SessionFile.load(
                self.config.app.session_file,
                allow_insecure=self.config.app.allow_insecure_auth_file,
            )
        except SessionFileError as e:
            logger.error("Unable to load session file: %s", e)
            return None
        except Exception as e:
            logger.error("Unexpected error loading session file: %s", e)
            return None

    def load_auth_token_file(self) -> Union[tuple[Path, str], tuple[None, None]]:
        """Attempts to load an auth token file."""
        paths = get_auth_token_file_paths(self.config)
        for path in paths:
            contents = self._do_load_auth_file(path)
            if contents:
                return path, contents
        logger.info(
            "No auth token file found. Searched in %s",
            {", ".join(str(p) for p in paths)},
        )
        return None, None

    def load_auth_file(self) -> tuple[Optional[Path], Optional[str]]:
        """Attempts to load an auth file."""
        paths = get_auth_file_paths(self.config)
        for path in paths:
            contents = self._do_load_auth_file(path)
            if contents:
                return path, contents
        logger.info(
            "No auth file found. Searched in %s", {", ".join(str(p) for p in paths)}
        )
        return None, None

    def _do_load_auth_file(self, file: Path) -> Optional[str]:
        """Attempts to read the contents of an auth (token) file.
        Returns None if the file does not exist or is not secure.
        """
        if not file.exists():
            return None
        if (
            not self.config.app.allow_insecure_auth_file
            and not file_has_secure_permissions(file)
        ):
            error(
                f"Auth file {file} must have {SECURE_PERMISSIONS_STR} permissions, has {oct(get_file_permissions(file))}. Refusing to load."
            )
            return None
        return file.read_text().strip()


def login(config: Config) -> ZabbixAPI:
    """Log in to the Zabbix API using credentials from the config.

    Returns the Zabbix API client object.
    """
    auth = Authenticator(config)
    client, _ = auth.login_with_any()
    return client


def logout(client: ZabbixAPI, config: Config) -> None:
    """Log out of the current Zabbix API session."""
    client.logout()
    if config.app.use_session_file:
        clear_auth_token_file(config)


def _parse_auth_file_contents(
    contents: Optional[str],
) -> tuple[Optional[str], Optional[str]]:
    """Parse the contents of an auth file.

    We store auth files in the format `username::secret`.
    """
    if contents:
        lines = contents.splitlines()
        if lines:
            line = lines[0].strip()
            username, _, secret = line.partition("::")
            return username, secret
    return None, None


def get_auth_file_paths(config: Optional[Config] = None) -> list[Path]:
    """Get all possible auth token file paths."""
    paths = [
        AUTH_FILE,
        AUTH_FILE_LEGACY,
    ]
    if config and config.app.auth_file not in paths:  # config has custom path
        paths.insert(0, config.app.auth_file)
    return paths


def get_auth_token_file_paths(config: Optional[Config] = None) -> list[Path]:
    """Get all possible auth token file paths."""
    paths = [
        AUTH_TOKEN_FILE,
        AUTH_TOKEN_FILE_LEGACY,
    ]
    if config and config.app.auth_token_file not in paths:  # config has custom path
        paths.insert(0, config.app.auth_token_file)
    return paths


def write_auth_token_file(
    username: str,
    auth_token: str,
    file: Path = AUTH_TOKEN_FILE,
    *,
    allow_insecure: bool = False,
) -> Path:
    """Write a username/auth token pair to the auth token file."""
    if not username or not auth_token:
        logger.error(
            "Cannot write auth token file without both a username (%s) and token (%s).",
            username,
            auth_token,
        )
        return file

    try:
        file.write_text(f"{username}::{auth_token}")
        logger.info("Wrote auth token file %s", file)
    except OSError as e:
        raise AuthTokenFileError(f"Unable to write auth token file {file}: {e}") from e

    # Ensure file has secure permissions if configured
    if not allow_insecure and not file_has_secure_permissions(file):
        try:
            file.chmod(SECURE_PERMISSIONS)
        except OSError as e:
            raise AuthTokenFileError(
                f"Unable to set secure permissions ({SECURE_PERMISSIONS_STR}) on {file} when saving auth token. "
                "Change permissions manually or delete the file."
            ) from e

    return file


def clear_auth_token_file(config: Optional[Config] = None) -> None:
    """Clear the contents of the auth token file.

    Attempts to clear both the new and the old auth token file locations.
    """
    for file in get_auth_token_file_paths(config):
        if not file.exists():
            continue
        try:
            file.write_text("")
            logger.debug("Cleared auth token file contents '%s'", file)
        except OSError as e:
            raise AuthTokenFileError(
                f"Unable to clear auth token file {file}: {e}"
            ) from e


def file_has_secure_permissions(file: Path) -> bool:
    """Check if a file has secure permissions.

    Always returns True on Windows.
    """
    if sys.platform == "win32":
        return True
    return get_file_permissions(file) == SECURE_PERMISSIONS


def set_file_secure_permissions(file: Path) -> None:
    """Set the permissions of a file to secure permissions."""
    try:
        file.chmod(SECURE_PERMISSIONS)
    except OSError as e:
        raise ZabbixCLIError(
            f"Unable to set secure permissions ({SECURE_PERMISSIONS_STR}) on {file}. Change permissions manually or delete the file."
        ) from e


def get_file_permissions(file: Path) -> int:
    """Get the 3 digit octal permissions of a file."""
    return file.stat().st_mode & 0o777



from __future__ import annotations

from typing import Optional

import typer

from zabbix_cli._v2_compat import ARGS_POSITIONAL
from zabbix_cli.app import Example
from zabbix_cli.app import app
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import info
from zabbix_cli.output.render import render_result
from zabbix_cli.pyzabbix.enums import InterfaceConnectionMode
from zabbix_cli.pyzabbix.enums import InterfaceType
from zabbix_cli.pyzabbix.enums import SNMPAuthProtocol
from zabbix_cli.pyzabbix.enums import SNMPPrivProtocol
from zabbix_cli.pyzabbix.enums import SNMPSecurityLevel
from zabbix_cli.utils.args import check_at_least_one_option_set

HELP_PANEL = "Host Interface"


@app.command(
    name="create_host_interface",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Create an SNMPv2 interface on host 'foo.example.com' with derived DNS name 'foo.example.com' (default)",
            "create_host_interface foo.example.com",
        ),
        Example(
            "Create an SNMPv2 interface on host 'foo.example.com' with IP connection",
            "create_host_interface foo.example.com --type snmp --ip 127.0.0.1",
        ),
        Example(
            "Create an SNMPv2 interface on host 'foo.example.com' with different DNS name",
            "create_host_interface foo.example.com --type snmp --dns snmp.example.com",
        ),
        Example(
            "Create an SNMPv2 interface on host 'foo' with both IP and DNS, using DNS as enabled address",
            "create_host_interface foo --type snmp --connection dns --dns snmp.example.com --ip 127.0.0.1",
        ),
        Example(
            "Create an SNMPv3 interface on host 'foo.example.com'",
            "create_host_interface foo.example.com --type snmp --snmp-version 3 --snmp-context-name mycontext --snmp-security-name myuser --snmp-security-level authpriv  --snmp-auth-protocol MD5 --snmp-auth-passphrase mypass --snmp-priv-protocol DES --snmp-priv-passphrase myprivpass",
        ),
        Example(
            "Create an Agent interface on host 'foo.example.com'",
            "create_host_interface foo.example.com --type agent",
        ),
    ],
)
def create_host_interface(
    ctx: typer.Context,
    hostname: str = typer.Argument(
        help="Name of host to create interface on.",
        show_default=False,
    ),
    connection: Optional[InterfaceConnectionMode] = typer.Option(
        None,
        "--connection",
        help="Interface connection mode. Required if both --ip and --dns are specified.",
        case_sensitive=False,
    ),
    type_: InterfaceType = typer.Option(
        InterfaceType.SNMP,
        "--type",
        help="Interface type. SNMP enables --snmp-* options.",
        case_sensitive=False,
    ),
    port: Optional[str] = typer.Option(
        None,
        "--port",
        help="Interface port. Defaults to 10050 for agent, 161 for SNMP, 623 for IPMI, and 12345 for JMX.",
    ),
    ip: Optional[str] = typer.Option(
        None,
        "--ip",
        help="IP address of interface.",
        show_default=False,
    ),
    dns: Optional[str] = typer.Option(
        None,
        "--dns",
        help="DNS address of interface.",
        show_default=False,
    ),
    default: bool = typer.Option(
        True, "--default/--no-default", help="Whether this is the default interface."
    ),
    snmp_version: int = typer.Option(
        2,
        "--snmp-version",
        help="SNMP version.",
        min=1,
        max=3,
        show_default=False,
    ),
    snmp_bulk: bool = typer.Option(
        True,
        "--snmp-bulk/--no-snmp-bulk",
        help="Use bulk SNMP requests.",
    ),
    snmp_community: str = typer.Option(
        "{$SNMP_COMMUNITY}",
        "--snmp-community",
        help="SNMPv{1,2} community.",
        show_default=False,
    ),
    snmp_max_repetitions: int = typer.Option(
        10,
        "--snmp-max-repetitions",
        help="Max repetitions for SNMPv{2,3} bulk requests.",
        min=1,
    ),
    snmp_security_name: Optional[str] = typer.Option(
        None,
        "--snmp-security-name",
        help="SNMPv3 security name.",
        show_default=False,
    ),
    snmp_context_name: Optional[str] = typer.Option(
        None,
        "--snmp-context-name",
        help="SNMPv3 context name.",
        show_default=False,
    ),
    snmp_security_level: Optional[SNMPSecurityLevel] = typer.Option(
        None,
        "--snmp-security-level",
        help="SNMPv3 security level.",
        show_default=False,
        case_sensitive=False,
    ),
    snmp_auth_protocol: Optional[SNMPAuthProtocol] = typer.Option(
        None,
        "--snmp-auth-protocol",
        help="SNMPv3 auth protocol (authNoPriv & authPriv).",
        show_default=False,
        case_sensitive=False,
    ),
    snmp_auth_passphrase: Optional[str] = typer.Option(
        None,
        "--snmp-auth-passphrase",
        help="SNMPv3 auth passphrase (authNoPriv & authPriv).",
        show_default=False,
    ),
    snmp_priv_protocol: Optional[SNMPPrivProtocol] = typer.Option(
        None,
        "--snmp-priv-protocol",
        help="SNMPv3 priv protocol (authPriv)",
        show_default=False,
        case_sensitive=False,
    ),
    snmp_priv_passphrase: Optional[str] = typer.Option(
        None,
        "--snmp-priv-passphrase",
        help="SNMPv3 priv passphrase (authPriv).",
        show_default=False,
    ),
    # V2-style positional args (deprecated)
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    """Create a host interface.

    Creates an SNMPv2 interface by default. Use --type to specify a different type.
    One of --dns and --ip is required. If both are specified, --connection is required.

    [b]NOTE:[/] Can only create secondary host interfaces for interfaces of types
    that already have a default interface. (API limitation)
    """
    from zabbix_cli.models import Result
    from zabbix_cli.pyzabbix.types import CreateHostInterfaceDetails

    # Handle V2 positional args (deprecated)
    if args:
        if len(args) != 6:  # 7 - 1 (hostname)
            exit_err(
                "create_host_interface takes exactly 6 positional arguments (deprecated)."
            )

        connection = InterfaceConnectionMode(args[0])
        type_ = InterfaceType(args[1])
        port = args[2]
        ip = args[3]
        dns = args[4]
        default = args[5] == "1"

    # Determine connection
    if not connection:
        if ip and dns:
            exit_err("Cannot specify both IP and DNS address without --connection.")
        elif ip:
            connection = InterfaceConnectionMode.IP
        else:
            connection = InterfaceConnectionMode.DNS
            if not dns:
                dns = hostname
    use_ip = connection == InterfaceConnectionMode.IP

    # Use default port for type if not specified
    if port is None:
        port = type_.get_port()

    host = app.state.client.get_host(hostname, select_interfaces=True)
    for interface in host.interfaces:
        if not interface.type == type_.as_api_value():
            continue
        if interface.main and interface.interfaceid:
            info(
                f"Host already has a default {type_} interface. New interface will be created as non-default."
            )
            default = False
            break
    else:
        # No default interface of this type found
        info(f"No default {type_} interface found. Setting new interface as default.")
        default = True

    details: Optional[CreateHostInterfaceDetails] = None
    if type_ == InterfaceType.SNMP:
        details = CreateHostInterfaceDetails(
            version=snmp_version,
            community=snmp_community,
            bulk=int(snmp_bulk),
        )

        if snmp_version > 1:
            details.max_repetitions = snmp_max_repetitions

        # V3-specific options
        if snmp_version == 3:
            if snmp_security_name:
                details.securityname = snmp_security_name
            if snmp_context_name:
                details.contextname = snmp_context_name
            if snmp_security_level:
                details.securitylevel = snmp_security_level.as_api_value()
            # authNoPriv and authPriv:
            if snmp_security_level != SNMPSecurityLevel.NO_AUTH_NO_PRIV:
                if snmp_auth_protocol:
                    details.authprotocol = snmp_auth_protocol.as_api_value()
                if snmp_auth_passphrase:
                    details.authpassphrase = snmp_auth_passphrase
                # authPriv:
                if snmp_security_level == SNMPSecurityLevel.AUTH_PRIV:
                    if snmp_priv_protocol:
                        details.privprotocol = snmp_priv_protocol.as_api_value()
                    if snmp_priv_passphrase:
                        details.privpassphrase = snmp_priv_passphrase

    ifaceid = app.state.client.create_host_interface(
        host=host,
        main=default,
        type=type_,
        use_ip=use_ip,
        ip=ip,
        dns=dns,
        port=str(port),
        details=details,
    )
    render_result(Result(message=f"Created host interface with ID {ifaceid}."))


@app.command(name="remove_host_interface", rich_help_panel=HELP_PANEL)
def remove_host_interface(
    ctx: typer.Context,
    interface_id: str = typer.Argument(
        help="ID of interface to remove.",
        show_default=False,
    ),
) -> None:
    """Remove a host interface."""
    from zabbix_cli.models import Result

    app.state.client.delete_host_interface(interface_id)
    render_result(Result(message=f"Removed host interface ({interface_id})."))


@app.command(name="show_host_interfaces", rich_help_panel=HELP_PANEL)
def show_host_interfaces(
    hostname_or_id: str = typer.Argument(
        help="Hostname or ID",
        show_default=False,
    ),
) -> None:
    """Show host interfaces."""
    from zabbix_cli.models import AggregateResult

    host = app.state.client.get_host(hostname_or_id, select_interfaces=True)
    render_result(AggregateResult(result=host.interfaces))


@app.command(
    name="update_host_interface",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Update the IP address of interface 123.",
            "update_host_interface 123 --ip 127.0.0.1",
        ),
        Example(
            "Change connection type of interface 123 to IP.",
            "update_host_interface 123 --connection ip",
        ),
        Example(
            "Change SNMP community of interface 234 to 'public'.",
            "update_host_interface 234 --snmp-community public",
        ),
    ],
)
def update_host_interface(
    ctx: typer.Context,
    interface_id: str = typer.Argument(
        help="ID of interface to update.",
        show_default=False,
    ),
    connection: Optional[InterfaceConnectionMode] = typer.Option(
        None,
        "--connection",
        help="Interface connection mode.",
        case_sensitive=False,
    ),
    port: Optional[str] = typer.Option(
        None,
        "--port",
        help="Interface port.",
    ),
    ip: Optional[str] = typer.Option(
        None,
        "--ip",
        help="IP address of interface.",
        show_default=False,
    ),
    dns: Optional[str] = typer.Option(
        None,
        "--dns",
        help="DNS address of interface.",
        show_default=False,
    ),
    default: bool = typer.Option(
        True, "--default/--no-default", help="Default interface."
    ),
    snmp_version: Optional[int] = typer.Option(
        None,
        "--snmp-version",
        help="SNMP version.",
        min=1,
        max=3,
        show_default=False,
    ),
    snmp_bulk: Optional[bool] = typer.Option(
        None,
        "--snmp-bulk/--no-snmp-bulk",
        help="Use bulk SNMP requests.",
        show_default=False,
    ),
    snmp_community: Optional[str] = typer.Option(
        None,
        "--snmp-community",
        help="SNMPv{1,2} community.",
        show_default=False,
    ),
    snmp_max_repetitions: Optional[int] = typer.Option(
        None,
        "--snmp-max-repetitions",
        help="Max repetitions for SNMPv{2,3} bulk requests.",
        min=1,
    ),
    snmp_security_name: Optional[str] = typer.Option(
        None,
        "--snmp-security-name",
        help="SNMPv3 security name.",
        show_default=False,
    ),
    snmp_context_name: Optional[str] = typer.Option(
        None,
        "--snmp-context-name",
        help="SNMPv3 context name.",
        show_default=False,
    ),
    snmp_security_level: Optional[SNMPSecurityLevel] = typer.Option(
        None,
        "--snmp-security-level",
        help="SNMPv3 security level.",
        show_default=False,
        case_sensitive=False,
    ),
    snmp_auth_protocol: Optional[SNMPAuthProtocol] = typer.Option(
        None,
        "--snmp-auth-protocol",
        help="SNMPv3 auth protocol (authNoPriv & authPriv).",
        show_default=False,
        case_sensitive=False,
    ),
    snmp_auth_passphrase: Optional[str] = typer.Option(
        None,
        "--snmp-auth-passphrase",
        help="SNMPv3 auth passphrase (authNoPriv & authPriv).",
        show_default=False,
    ),
    snmp_priv_protocol: Optional[SNMPPrivProtocol] = typer.Option(
        None,
        "--snmp-priv-protocol",
        help="SNMPv3 priv protocol (authPriv)",
        show_default=False,
        case_sensitive=False,
    ),
    snmp_priv_passphrase: Optional[str] = typer.Option(
        None,
        "--snmp-priv-passphrase",
        help="SNMPv3 priv passphrase (authPriv).",
        show_default=False,
    ),
) -> None:
    """Update a host interface.

    Host interface type cannot be changed.
    """
    from zabbix_cli.models import Result
    from zabbix_cli.pyzabbix.types import UpdateHostInterfaceDetails

    check_at_least_one_option_set(ctx)

    interface = app.state.client.get_hostinterface(interface_id)

    details = UpdateHostInterfaceDetails(
        version=snmp_version,
        community=snmp_community,
        bulk=int(snmp_bulk) if snmp_bulk is not None else None,
        max_repetitions=snmp_max_repetitions,
        securityname=snmp_security_name,
        contextname=snmp_context_name,
        securitylevel=snmp_security_level.as_api_value()
        if snmp_security_level
        else None,
        authprotocol=snmp_auth_protocol.as_api_value() if snmp_auth_protocol else None,
        authpassphrase=snmp_auth_passphrase,
        privprotocol=snmp_priv_protocol.as_api_value() if snmp_priv_protocol else None,
        privpassphrase=snmp_priv_passphrase,
    )

    if connection:
        if connection == InterfaceConnectionMode.IP:
            use_ip = True
        elif connection == InterfaceConnectionMode.DNS:
            use_ip = False
    else:
        use_ip = None

    app.state.client.update_host_interface(
        interface,
        main=default,
        use_ip=use_ip,
        ip=ip,
        dns=dns,
        port=port,
        details=details,
    )
    render_result(Result(message=f"Updated host interface ({interface_id})."))



"""Compatibility functions going from Zabbix-CLI v2 to v3.

The functions in this module are intended to ease the transition by
providing fallbacks to deprecated functionality in Zabbix-CLI v2.
"""

from __future__ import annotations

import os
import shlex
from pathlib import Path
from typing import Optional

import typer
from click.core import CommandCollection
from click.core import Group

CONFIG_FILENAME = "zabbix-cli.conf"
CONFIG_FIXED_NAME = "zabbix-cli.fixed.conf"

# Config file locations
CONFIG_DEFAULT_DIR = "/usr/share/zabbix-cli"
CONFIG_SYSTEM_DIR = "/etc/zabbix-cli"
CONFIG_USER_DIR = os.path.expanduser("~/.zabbix-cli")

# Any item will overwrite values from the previous (NYI)
CONFIG_PRIORITY = tuple(
    Path(os.path.join(d, f))
    for d, f in (
        (CONFIG_DEFAULT_DIR, CONFIG_FIXED_NAME),
        (CONFIG_SYSTEM_DIR, CONFIG_FIXED_NAME),
        (CONFIG_USER_DIR, CONFIG_FILENAME),
        (CONFIG_SYSTEM_DIR, CONFIG_FILENAME),
        (CONFIG_DEFAULT_DIR, CONFIG_FILENAME),
    )
)


AUTH_FILE = Path.home() / ".zabbix-cli_auth"
AUTH_TOKEN_FILE = Path.home() / ".zabbix-cli_auth_token"


def run_command_from_option(ctx: typer.Context, command: str) -> None:
    """Runs a command via old-style --command/-C option."""
    from zabbix_cli.output.console import error
    from zabbix_cli.output.console import exit_err
    from zabbix_cli.output.console import warning

    warning(
        "The [i]--command/-C[/] option is deprecated and will be removed in a future release. "
        "Invoke command directly instead."
    )
    if not isinstance(ctx.command, (CommandCollection, Group)):
        exit_err(  # TODO: find out if this could ever happen?
            f"Cannot run command {command!r}. Ensure it is a valid command and try again."
        )

    parts = shlex.split(command, comments=True)
    if not parts:
        exit_err(
            f"Command {command!r} is empty. Ensure it is a valid command and try again."
        )

    try:
        with ctx.command.make_context(None, parts, parent=ctx) as new_ctx:
            ctx.command.invoke(new_ctx)
    except typer.Exit:
        pass
    except Exception as e:
        error(
            f"Command {command!r} failed with error: {e}. Try re-running without --command."
        )
        raise


def args_callback(
    ctx: typer.Context, value: Optional[list[str]]
) -> Optional[list[str]]:
    if ctx.resilient_parsing:
        return  # for auto-completion
    if value:
        from zabbix_cli.output.console import warning

        warning(
            f"Detected deprecated positional arguments {value}. Use options instead."
        )
    # NOTE: Must NEVER return None. The "fix" in Typer 0.10.0 for None defaults
    # somehow broke the parsing of callback values by causing values returned by
    # callbacks to be passed to the internal converter, which then fails
    # because it expects a list but gets None.
    # https://github.com/tiangolo/typer/pull/664
    # https://github.com/tiangolo/typer/blob/142422a14ca4c6a8ad579e9bd0fd0728364d86e3/typer/main.py#L639
    return value or []


ARGS_POSITIONAL = typer.Argument(
    None,
    help="DEPRECATED: V2-style positional arguments.",
    show_default=False,
    hidden=True,
    callback=args_callback,
)



from __future__ import annotations

from typing import Any
from typing import Optional

import click
import typer
from click.types import ParamType
from typer.core import TyperGroup

from zabbix_cli.logs import logger


def get_limit_option(
    limit: Optional[int] = 0,
    resource: str = "results",
    long_option: str = "--limit",
    short_option: str = "-n",
) -> Any:  # TODO: Can we type this better?
    """Limit option factory."""
    return typer.Option(
        limit,
        long_option,
        short_option,
        help=f"Limit the number of {resource}. 0 to show all.",
    )


OPTION_LIMIT = get_limit_option(0)

ARG_TEMPLATE_NAMES_OR_IDS = typer.Argument(
    help="Template names or IDs. Comma-separated. Supports wildcards.",
    show_default=False,
)
ARG_HOSTNAMES_OR_IDS = typer.Argument(
    help="Hostnames or IDs. Comma-separated. Supports wildcards.",
    show_default=False,
)
ARG_GROUP_NAMES_OR_IDS = typer.Argument(
    help="Host/template group names or IDs. Comma-separated. Supports wildcards.",
    show_default=False,
)


class CommandParam(ParamType):
    """Command param type that resolves into a click Command."""

    name = "command"

    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> click.Command:
        if not value:
            self.fail("Missing command.", param, ctx)
        if not ctx:
            self.fail("No context.", param, ctx)
        root_ctx = ctx.find_root()
        root_command = root_ctx.command

        if not isinstance(root_command, TyperGroup):
            logger.error(
                "Root context of %s is not a TyperGroup, unable to show help",
                root_command,
            )
            self.fail(f"Unable to show help for '{value}'")

        cmd = root_command.get_command(root_ctx, value)
        if not cmd:
            self.fail(f"Command '{value}' not found.")
        return cmd



from __future__ import annotations

import pytest
from zabbix_cli.update import PyInstallerUpdater


@pytest.mark.parametrize(
    "os, arch, version, expect_info",
    [
        ("linux", "x86_64", "1.2.3", "1.2.3-linux-x86_64"),
        ("linux", "arm64", "1.2.3", "1.2.3-linux-arm64"),
        ("linux", "armv7l", "1.2.3", "1.2.3-linux-armv7l"),
        ("darwin", "x86_64", "1.2.3", "1.2.3-macos-x86_64"),
        ("darwin", "arm64", "1.2.3", "1.2.3-macos-arm64"),
        ("win32", "x86_64", "1.2.3", "1.2.3-win-x86_64.exe"),
    ],
)
def test_pyinstaller_updater_get_url(
    os: str, arch: str, version: str, expect_info: str
):
    BASE_URL = (
        "https://github.com/unioslo/zabbix-cli/releases/latest/download/zabbix-cli"
    )
    expect_url = f"{BASE_URL}-{expect_info}"

    url = PyInstallerUpdater.get_url(os, arch, version)
    assert url == expect_url



"""Generate the code reference pages and navigation."""

from __future__ import annotations

import sys
from pathlib import Path
from typing import Any

import jinja2
import yaml  # type: ignore
from sanitize_filename import sanitize
from zabbix_cli.app import app

sys.path.append(Path(__file__).parent.as_posix())
sys.path.append(Path(__file__).parent.parent.parent.as_posix())

from common import COMMANDS_DIR  # noqa
from common import DATA_DIR  # noqa
from common import TEMPLATES_DIR  # noqa
from utils.commands import CommandSummary  # noqa: E402
from utils.commands import get_app_commands  # noqa: E402


def gen_command_list(commands: list[CommandSummary]) -> None:
    """Generates a YAML file with a list of the names of all commands."""
    command_names = [c.name for c in commands]
    with open(DATA_DIR / "commandlist.yaml", "w") as f:
        yaml.dump(command_names, f, sort_keys=False)


def gen_category_command_map(commands: list[CommandSummary]) -> None:
    """Generates a YAML file with all categories and detailed information
    about their respective commands.
    """
    categories: dict[str, list[dict[str, Any]]] = {}
    for command in commands:
        category = command.category or ""
        if category not in categories:
            categories[category] = []
        cmd_dict = command.model_dump(mode="json")
        # cmd_dict["usage"] = command.usage
        categories[category].append(cmd_dict)

    with open(DATA_DIR / "commands.yaml", "w") as f:
        yaml.dump(categories, f, sort_keys=True)


def gen_category_pages(commands: list[CommandSummary]) -> None:
    """Renders markdown pages for each category with detailed information
    about each command.
    """
    categories: dict[str, list[CommandSummary]] = {}
    for command in commands:
        if command.hidden:
            continue
        category = command.category or command.name
        if category not in categories:
            categories[category] = []
        categories[category].append(command)

    loader = jinja2.FileSystemLoader(searchpath=TEMPLATES_DIR)
    env = jinja2.Environment(loader=loader)

    # Render each individual command page
    pages = {}  # type: dict[str, str] # {category: filename}
    for category_name, cmds in categories.items():
        template = env.get_template("category.md.j2")
        filename = sanitize(category_name.replace(" ", "_"))
        filepath = COMMANDS_DIR / f"{filename}.md"
        with open(filepath, "w") as f:
            f.write(template.render(category=category_name, commands=cmds))
        pages[category_name] = filename


def main() -> None:
    commands = get_app_commands(app)
    gen_category_command_map(commands)
    gen_command_list(commands)
    gen_category_pages(commands)


if __name__ == "__main__":
    main()



"""Script that runs various CLI commands and collects the result for use
in the documentation.

The commands are run in a limited environment (no color, limited width) to
make the output more readable in the documentation.
"""

from __future__ import annotations

import os
import subprocess
import sys
from pathlib import Path
from typing import NamedTuple
from typing import Optional
from typing import Protocol

try:
    import tomllib
except ImportError:
    import tomli as tomllib
import tomli_w

sys.path.append(Path(__file__).parent.as_posix())


from common import DATA_DIR  # noqa: I001
from common import add_path_placeholders

# Set up environment variables for the CLI
env = os.environ.copy()
env["LINES"] = "40"
env["COLUMNS"] = "90"  # limit width so it looks nicer in MD code blocks
env["TERM"] = "dumb"  # disable color output (color codes mangle it)


class CommandCallback(Protocol):
    def __call__(self, output: str) -> str: ...


class Command(NamedTuple):
    command: list[str]
    filename: str
    callback: Optional[CommandCallback] = None


def add_config_bogus_defaults(output: str) -> str:
    """Give bogus defaults to certain config values."""
    config = tomllib.loads(output)
    # TODO: replace local username with a default value
    out = tomli_w.dumps(config)
    out = add_path_placeholders(out)
    return out


COMMAND_HELP = Command(["zabbix-cli", "--help"], "help.txt")
COMMAND_SAMPLE_CONFIG = Command(
    ["zabbix-cli", "sample_config"],
    "sample_config.toml",
    callback=add_config_bogus_defaults,
)

# List of commands to run
COMMANDS = [
    COMMAND_HELP,
    COMMAND_SAMPLE_CONFIG,
]


def main() -> None:
    """Run the commands and save the output to files."""
    for cmd in COMMANDS:
        output = subprocess.check_output(cmd.command, env=env).decode("utf-8")
        if cmd.callback:
            output = cmd.callback(output)

        with open(DATA_DIR / cmd.filename, "w") as f:
            f.write(output)


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Optional
from typing import Union

from strenum import StrEnum

from zabbix_cli.dirs import CONFIG_DIR
from zabbix_cli.dirs import DATA_DIR
from zabbix_cli.dirs import LOGS_DIR
from zabbix_cli.dirs import SITE_CONFIG_DIR

logger = logging.getLogger("zabbix_cli.config")

# Config file basename
CONFIG_FILENAME = "zabbix-cli.toml"
DEFAULT_CONFIG_FILE = CONFIG_DIR / CONFIG_FILENAME


CONFIG_PRIORITY = (
    Path() / CONFIG_FILENAME,  # current directory
    DEFAULT_CONFIG_FILE,  # local config directory
    SITE_CONFIG_DIR / CONFIG_FILENAME,  # system config directory
)


# File path defaults
AUTH_TOKEN_FILE = DATA_DIR / ".zabbix-cli_auth_token"
"""Path to file containing API session token."""

AUTH_FILE = DATA_DIR / ".zabbix-cli_auth"
"""Path to file containing user credentials."""

SESSION_FILE = DATA_DIR / ".zabbix-cli_session.json"
"""Path to JSON file containing API session IDs."""

HISTORY_FILE = DATA_DIR / "history"
"""Path to file containing REPL history."""

LOG_FILE = LOGS_DIR / "zabbix-cli.log"


# Environment variable names
class ConfigEnvVars:
    API_TOKEN = "ZABBIX_API_TOKEN"
    PASSWORD = "ZABBIX_PASSWORD"
    URL = "ZABBIX_URL"
    USERNAME = "ZABBIX_USERNAME"


class OutputFormat(StrEnum):
    # CSV = "csv"
    JSON = "json"
    """JSON-serialized output."""

    TABLE = "table"
    """Rich terminal table output."""


class SecretMode(StrEnum):
    """Mode for serializing secrets."""

    HIDE = "hide"
    MASK = "masked"
    PLAIN = "plain"

    _DEFAULT = MASK

    @classmethod
    def from_context(
        cls, context: Optional[Union[dict[str, Any], Callable[..., dict[str, Any]]]]
    ) -> SecretMode:
        """Get the secret mode from a serialization context."""
        if isinstance(context, dict) and (ctx := context.get("secrets")) is not None:
            # Support for old-style context values (true/false)
            # as well as new context values (enum values)
            try:
                if isinstance(ctx, SecretMode):
                    return ctx
                elif ctx is True:
                    return cls.PLAIN
                elif ctx is False:
                    return cls.MASK
                else:
                    return cls(str(ctx).lower())
            except ValueError:
                logger.warning(
                    "Got invalid secret mode from context %s: %s", context, ctx
                )
        return cls._DEFAULT


class BulkRunnerMode(StrEnum):
    """Mode of operation for BulkRunner."""

    STRICT = "strict"
    """Stop on first error."""

    CONTINUE = "continue"
    """Continue on errors, report at end."""

    SKIP = "skip"
    """Skip lines with errors. No reporting."""



from __future__ import annotations

import functools
from typing import TYPE_CHECKING
from typing import Any
from typing import NoReturn
from typing import Optional
from typing import Protocol
from typing import runtime_checkable

if TYPE_CHECKING:
    from httpx import ConnectError
    from httpx import RequestError
    from httpx import Response as HTTPResponse
    from pydantic import BaseModel
    from pydantic import ValidationError

    from zabbix_cli.pyzabbix.types import ParamsType
    from zabbix_cli.pyzabbix.types import ZabbixAPIResponse


class ZabbixCLIError(Exception):
    """Base exception class for ZabbixCLI exceptions."""


class ZabbixCLIFileError(ZabbixCLIError, OSError):
    """Errors related to reading/writing files."""


class ZabbixCLIFileNotFoundError(ZabbixCLIError, FileNotFoundError):
    """Errors related to reading/writing files."""


class ConfigError(ZabbixCLIError):
    """Error with configuration file."""


class ConfigExistsError(ConfigError):
    """Configuration file already exists."""


class ConfigOptionNotFound(ConfigError):
    """Configuration option is missing from the loaded config."""


class CommandFileError(ZabbixCLIError):
    """Error running bulk commands from a file."""


class AuthError(ZabbixCLIError):
    """Base class for all authentication errors."""

    # NOTE: We might still run into the problem of expired tokens, which won't raise
    # this type of error, but instead raise a ZabbixAPIRequestError.
    # We should probably handle that in the client and raise the appropriate exception


class SessionFileError(AuthError):
    """Session file error."""


class SessionFileNotFoundError(SessionFileError, FileNotFoundError):
    """Session file does not exist."""


class SessionFilePermissionsError(SessionFileError):
    """Session file has incorrect permissions."""


class AuthTokenFileError(AuthError):
    """Auth token file error."""


class AuthTokenError(AuthError):
    """Auth token (not file) error."""


class PluginError(ZabbixCLIError):
    """Plugin error."""


class PluginConfigError(PluginError, ConfigError):
    """Plugin configuration error."""


class PluginConfigTypeError(PluginConfigError, TypeError):
    """Plugin configuration type error."""


class PluginLoadError(PluginError):
    """Error loading a plugin."""

    msg = "Error loading plugin '{plugin_name}'"

    def __init__(self, plugin_name: str, plugin_config: BaseModel | None) -> None:
        self.plugin_name = plugin_name
        self.plugin_config = plugin_config
        super().__init__(self.msg.format(plugin_name=plugin_name))


class PluginPostImportError(PluginLoadError):
    """Error running post-import configuration for a plugin."""

    msg = "Error running post-import configuration for plugin '{plugin_name}'"


class ZabbixAPIException(ZabbixCLIError):
    # Extracted from pyzabbix, hence *Exception suffix instead of *Error
    """Base exception class for Zabbix API exceptions."""

    def reason(self) -> str:
        return ""


class ZabbixAPIRequestError(ZabbixAPIException):
    """Zabbix API response error."""

    def __init__(
        self,
        *args: Any,
        params: Optional[ParamsType] = None,
        api_response: Optional[ZabbixAPIResponse] = None,
        response: Optional[HTTPResponse] = None,
    ) -> None:
        super().__init__(*args)
        self.params = params
        self.api_response = api_response
        self.response = response

    def reason(self) -> str:
        if self.api_response and self.api_response.error:
            reason = (
                f"({self.api_response.error.code}) {self.api_response.error.message}"
            )
            if self.api_response.error.data:
                reason += f" {self.api_response.error.data}"
        elif self.response and self.response.text:
            reason = self.response.text
        else:
            reason = str(self)
        return reason


class ZabbixAPITokenExpiredError(ZabbixAPIRequestError, AuthError):
    """Zabbix API token expired error."""


class ZabbixAPINotAuthorizedError(ZabbixAPIRequestError):
    """Zabbix API not authorized error."""


class ZabbixAPIResponseParsingError(ZabbixAPIRequestError):
    """Zabbix API request error."""


class ZabbixAPISessionExpired(ZabbixAPIRequestError):
    """Zabbix API session expired."""


class ZabbixAPICallError(ZabbixAPIException):
    """Zabbix API request error."""

    def __str__(self) -> str:
        msg = super().__str__()
        if self.__cause__ and isinstance(self.__cause__, ZabbixAPIRequestError):
            msg = f"{msg}: {self.__cause__.reason()}"
        return msg


class ZabbixAPILoginError(ZabbixAPICallError, AuthError):
    """Zabbix API login error."""


class ZabbixAPILogoutError(ZabbixAPICallError, AuthError):
    """Zabbix API logout error."""


class ZabbixNotFoundError(ZabbixAPICallError):
    """A Zabbix API resource was not found."""


class Exiter(Protocol):
    """Protocol class for exit function that can be passed to an
    exception handler function.

    See Also:
    --------
    [zabbix_cli.exceptions.HandleFunc][]
    """

    def __call__(
        self,
        message: str,
        code: int = ...,
        exception: Optional[Exception] = ...,
        **kwargs: Any,
    ) -> NoReturn: ...


@runtime_checkable
class HandleFunc(Protocol):
    """Interface for exception handler functions.

    They take any exception and an Exiter function as the arguments
    and exit with the appropriate message after running any necessary
    cleanup and/or logging.
    """

    def __call__(self, e: Any) -> NoReturn: ...


def get_cause_args(e: Optional[BaseException]) -> list[str]:
    """Retrieves all args as strings from all exceptions in the cause chain.
    Flattens the args into a single list.
    """
    args: list[str] = []
    while e:
        args.extend(get_exc_args(e))
        e = e.__cause__
    return args


def get_exc_args(e: BaseException) -> list[str]:
    """Returns the error message as a string."""
    args: list[str] = [str(arg) for arg in e.args]
    if isinstance(e, ZabbixAPIRequestError):
        args.append(e.reason())
    return args


def handle_notraceback(e: Exception) -> NoReturn:
    """Handles an exception with no traceback in console.
    The exception is logged with a traceback in the log file.
    """
    get_exit_err()(str(e), exception=e, exc_info=True)


def handle_validation_error(e: ValidationError) -> NoReturn:
    """Handles a Pydantic validation error."""
    # TODO: Use some very primitive heuristics to determine whether or not
    # the error is from an API response or somewhere else
    get_exit_err()(str(e), exception=e, exc_info=True)


def _fmt_request_error(e: RequestError, exc_type: str, reason: str) -> str:
    method = e.request.method
    url = e.request.url
    return f"{exc_type}: {method} {url} - {reason}"


def handle_connect_error(e: ConnectError) -> NoReturn:
    """Handles an httpx ConnectError."""
    # Simple heuristic here to determine cause
    if e.args and "connection refused" in str(e.args[0]).casefold():
        reason = "Connection refused"
    else:
        reason = str(e)
    msg = _fmt_request_error(e, "Connection error", reason)
    get_exit_err()(msg, exception=e, exc_info=False)


def handle_zabbix_api_exception(e: ZabbixAPIException) -> NoReturn:
    """Handles a ZabbixAPIException."""
    from zabbix_cli.state import get_state

    state = get_state()
    # If we have a stale auth token, we need to clear it.
    if (
        state.is_config_loaded
        and state.config.app.use_session_file
        and any("re-login" in arg for arg in get_cause_args(e))
    ):
        from zabbix_cli.auth import clear_auth_token_file
        from zabbix_cli.output.console import error

        # Clear token file and from the config object
        error("Auth token expired. You must re-authenticate.")
        clear_auth_token_file(state.config)
        if state.repl:  # Hack: run login flow again in REPL
            state.login()
        # NOTE: ideally we automatically re-run the command here, but that's
        # VERY hacky and could lead to unexpected behavior.
        raise SystemExit(1)  # Exit without a message
    else:
        # TODO: extract the reason for the error from the exception here
        # and add it to the message.
        handle_notraceback(e)


def get_exception_handler(type_: type[Exception]) -> Optional[HandleFunc]:
    """Returns the exception handler for the given exception type."""
    from httpx import ConnectError
    from pydantic import ValidationError

    # Defined inline for performance reasons (httpx and pydantic imports)
    EXC_HANDLERS: dict[type[Exception], HandleFunc] = {
        # ZabbixAPICallError: handle_zabbix_api_call_error,  # NOTE: use different strategy for this?
        ZabbixAPIException: handle_zabbix_api_exception,  # NOTE: use different strategy for this?
        ZabbixCLIError: handle_notraceback,
        ValidationError: handle_validation_error,
        ConnectError: handle_connect_error,
        ConfigError: handle_notraceback,  # NOTE: can we remove this? subclass of ZabbixCLIError
    }
    """Mapping of exception types to exception handling strategies."""

    handler = EXC_HANDLERS.get(type_, None)
    if handler:
        return handler
    if type_.__bases__:
        for base in type_.__bases__:
            handler = get_exception_handler(base)
            if handler:
                return handler
    return None


def handle_exception(e: Exception) -> NoReturn:
    """Handles an exception and exits with the appropriate message."""
    handler = get_exception_handler(type(e))
    if not handler:
        raise e
    handler(e)


@functools.lru_cache(maxsize=1)
def get_exit_err() -> Exiter:
    """Cached lazy-import of `zabbix_cli.output.console.exit_err`.
    Avoids circular imports. Because we can "exit" multiple times in the
    REPL, it's arguably worth caching the import this way.
    """
    from zabbix_cli.output.console import exit_err as _exit_err

    return _exit_err



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Optional

from pydantic import Field
from pydantic import computed_field

from zabbix_cli.models import TableRenderable
from zabbix_cli.pyzabbix.types import Item

if TYPE_CHECKING:
    from zabbix_cli.models import ColsRowsType
    from zabbix_cli.models import RowsType


class UngroupedItem(TableRenderable):
    itemid: str
    name: Optional[str]
    key: Optional[str]
    lastvalue: Optional[str]
    host: Optional[str]

    @classmethod
    def from_item(cls, item: Item) -> UngroupedItem:
        return cls(
            itemid=item.itemid,
            name=item.name,
            key=item.key,
            lastvalue=item.lastvalue,
            host=item.hosts[0].host if item.hosts else None,
        )

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["Item ID", "Name", "Key", "Last value", "Host"]
        rows: RowsType = [
            [
                self.itemid,
                self.name or "",
                self.key or "",
                self.lastvalue or "",
                self.host or "",
            ]
        ]
        return cols, rows


class ItemResult(Item):
    """Alternate rendering of Item."""

    grouped: bool = Field(False, exclude=True)

    @classmethod
    def from_item(cls, item: Item) -> ItemResult:
        return cls.model_validate(item, from_attributes=True)

    @computed_field
    @property
    def host(self) -> str:
        """LEGACY: serialize list of hosts as newline-delimited string."""
        return "\n".join(h.host for h in self.hosts)

    def __cols_rows__(self) -> ColsRowsType:
        # As long as we include the "host" computed field, we need to
        # override the __cols_rows__ method.
        cols = ["Name", "Key", "Last value", "Hosts"]
        rows: RowsType = [
            [
                self.name or "",
                self.key or "",
                self.lastvalue or "",
                "\n".join(h.host for h in self.hosts),
            ],
        ]
        if self.grouped:
            cols.insert(0, "Item ID")
            rows[0].insert(0, self.itemid)
        return cols, rows


def group_items(items: list[Item]) -> list[ItemResult]:
    """Group items by key+lastvalue.

    Keeps first item for each key+lastvalue pair, and adds hosts from
    duplicate items to the first item.


    Example:
    ```py
        # Given the following items:
        >>> items = [
            Item(itemid="1", key="foo", lastvalue="bar", hosts=[Host(hostid="1")]),
            Item(itemid="2", key="foo", lastvalue="bar", hosts=[Host(hostid="2")]),
            Item(itemid="3", key="baz", lastvalue="baz", hosts=[Host(hostid="3")]),
            Item(itemid="4", key="baz", lastvalue="baz", hosts=[Host(hostid="4")]),
        ]
        >>> group_items(items)
        [
            Item(itemid="1", key="foo", lastvalue="bar", hosts=[Host(hostid="1"), Host(hostid="2")]),
            Item(itemid="3", key="baz", lastvalue="baz", hosts=[Host(hostid="3"), Host(hostid="4")]),
        ]
        # Hosts from items 2 and 4 were added to item 1 and 3, respectively.
    ```
    """
    from zabbix_cli.commands.results.item import ItemResult

    item_map: dict[str, ItemResult] = {}

    for item in items:
        if not item.name or not item.lastvalue or not item.key or not item.hosts:
            continue
        key = item.key + item.lastvalue
        for host in item.hosts:
            if key in item_map:
                item_map[key].hosts.append(host)
            else:
                res = ItemResult.from_item(item)
                res.grouped = True
                item_map[key] = res
    return list(item_map.values())



from __future__ import annotations

from inline_snapshot import snapshot
from zabbix_cli.bulk import BulkRunnerMode
from zabbix_cli.commands.cli_configwizard import get_enum_attr_docs
from zabbix_cli.commands.cli_configwizard import is_enum_type
from zabbix_cli.config.constants import OutputFormat


def test_get_enum_attr_docs():
    """Test docstring retrieval for enum attributes."""
    assert get_enum_attr_docs(BulkRunnerMode) == snapshot(
        {
            BulkRunnerMode.STRICT: "Stop on first error.",
            BulkRunnerMode.CONTINUE: "Continue on errors, report at end.",
            BulkRunnerMode.SKIP: "Skip lines with errors. No reporting.",
        }
    )
    assert get_enum_attr_docs(OutputFormat) == snapshot(
        {
            OutputFormat.JSON: "JSON-serialized output.",
            OutputFormat.TABLE: "Rich terminal table output.",
        }
    )


def test_is_enum():
    """Test if an object is an enum type."""

    assert is_enum_type(BulkRunnerMode)
    assert is_enum_type(OutputFormat)
    assert not is_enum_type(str)
    assert not is_enum_type(int)
    assert not is_enum_type(list)
    assert not is_enum_type(dict)
    assert not is_enum_type(tuple)

    # Instances
    assert not is_enum_type(BulkRunnerMode.STRICT)
    assert not is_enum_type(OutputFormat.JSON)
    assert not is_enum_type("test")
    assert not is_enum_type(123)
    assert not is_enum_type(["test"])
    assert not is_enum_type({"test": "test"})
    assert not is_enum_type((1, 2, 3))



from __future__ import annotations

import logging

import pytest
from inline_snapshot import snapshot
from pydantic import BaseModel
from pydantic import Field
from pytest import LogCaptureFixture
from zabbix_cli.models import MetaKey
from zabbix_cli.models import TableRenderable


@pytest.mark.parametrize(
    "header, expect",
    [
        pytest.param(None, "Foo", id="None is default"),
        pytest.param("", "Foo", id="Empty string is default"),
        ("Foo Header", "Foo Header"),
    ],
)
def test_table_renderable_metakey_header(header: str, expect: str) -> None:
    class TestTableRenderable(TableRenderable):
        foo: str = Field(..., json_schema_extra={MetaKey.HEADER: header})

    t = TestTableRenderable(foo="bar")
    assert t.__cols__() == [expect]
    assert t.__rows__() == [["bar"]]
    assert t.__cols_rows__() == ([expect], [["bar"]])


@pytest.mark.parametrize(
    "content, join_char, expect",
    [
        (["a", "b", "c"], ",", ["a,b,c"]),
        (["a", "b", "c"], "|", ["a|b|c"]),
        (["a", "b", "c"], " ", ["a b c"]),
        (["a", "b", "c"], "", ["abc"]),
        # Test empty list
        ([], ",", [""]),
        ([], "|", [""]),
        ([], " ", [""]),
        ([], "", [""]),
    ],
)
def test_table_renderable_metakey_join_char(
    content: list[str], join_char: str, expect: str
) -> None:
    class TestTableRenderable(TableRenderable):
        foo: list[str] = Field(..., json_schema_extra={MetaKey.JOIN_CHAR: join_char})

    t = TestTableRenderable(foo=content)
    assert t.__rows__() == [expect]


def test_all_metakeys() -> None:
    class TestTableRenderable(TableRenderable):
        foo: list[str] = Field(
            ...,
            json_schema_extra={MetaKey.JOIN_CHAR: "|", MetaKey.HEADER: "Foo Header"},
        )

    t = TestTableRenderable(foo=["foo", "bar"])
    assert t.__cols__() == ["Foo Header"]
    assert t.__rows__() == [["foo|bar"]]
    assert t.__cols_rows__() == (["Foo Header"], [["foo|bar"]])


def test_rows_with_unknown_base_model(caplog: LogCaptureFixture) -> None:
    """Test that we log when we try to render a BaseModel
    instance that does not inherit from TableRenderable.
    """

    class FooModel(BaseModel):
        foo: str
        bar: int
        baz: float
        qux: list[str]

    class TestTableRenderable(TableRenderable):
        foo: FooModel

    t = TestTableRenderable(foo=FooModel(foo="foo", bar=1, baz=1.0, qux=["a", "b"]))

    caplog.set_level(logging.WARNING)

    # Non-TableRenderable models are rendered as JSON
    assert t.__rows__() == snapshot(
        [
            [
                """\
{
  "foo": "foo",
  "bar": 1,
  "baz": 1.0,
  "qux": [
    "a",
    "b"
  ]
}\
"""
            ]
        ]
    )

    # Check that we logged info on what happened and how we got there
    assert caplog.record_tuples == snapshot(
        [("zabbix_cli", 30, "Cannot render FooModel as a table.")]
    )
    record = caplog.records[0]
    assert record.funcName == "__rows__"
    assert record.stack_info is not None
    assert "test_rows_with_unknown_base_model" in record.stack_info



from __future__ import annotations

from typing import Optional

import typer

from zabbix_cli.app import Example
from zabbix_cli.app import app
from zabbix_cli.commands.common.args import OPTION_LIMIT
from zabbix_cli.output.console import error
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import info
from zabbix_cli.output.console import success
from zabbix_cli.output.formatting.grammar import pluralize as p
from zabbix_cli.output.render import render_result
from zabbix_cli.pyzabbix.enums import UsergroupPermission
from zabbix_cli.utils.args import parse_hostgroups_arg
from zabbix_cli.utils.args import parse_hosts_arg
from zabbix_cli.utils.args import parse_list_arg

HELP_PANEL = "Host Group"


@app.command(
    "add_host_to_hostgroup",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Add a host to a host group",
            "add_host_to_hostgroup 'My host' 'My host group'",
        ),
        Example(
            "Add multiple hosts to a host group",
            "add_host_to_hostgroup 'host1,host2' 'My host group'",
        ),
        Example(
            "Add multiple hosts to multiple host groups",
            "add_host_to_hostgroup 'host1,host2' 'My host group,Another group'",
        ),
    ],
)
def add_host_to_hostgroup(
    ctx: typer.Context,
    hostnames_or_ids: str = typer.Argument(
        help="Host names or IDs. Comma-separated. Supports wildcards.",
        metavar="HOSTS",
        show_default=False,
    ),
    hostgroups: str = typer.Argument(
        help="Host group names or IDs. Comma-separated. Supports wildcards.",
        metavar="HOSTGROUPS",
        show_default=False,
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Preview changes",
    ),
) -> None:
    """Add hosts to host groups.

    Host name and group arguments are interpreted as IDs if they are numeric.
    """
    import itertools

    from zabbix_cli.commands.results.hostgroup import AddHostsToHostGroup
    from zabbix_cli.models import AggregateResult

    hosts = parse_hosts_arg(app, hostnames_or_ids)
    hgs = parse_hostgroups_arg(app, hostgroups, select_hosts=True)
    if not dryrun:
        with app.status("Adding hosts to host groups..."):
            app.state.client.add_hosts_to_hostgroups(hosts, hgs)

    result: list[AddHostsToHostGroup] = []
    for hg in hgs:
        r = AddHostsToHostGroup.from_result(hosts, hg)
        if not r.hosts:
            continue
        result.append(r)

    total_hosts = len(set(itertools.chain.from_iterable((r.hosts) for r in result)))
    total_hgs = len(result)

    if not total_hosts:
        exit_err("No hosts to add.")

    render_result(AggregateResult(result=result))
    base_msg = f"{p('host', total_hosts)} to {p('host group', total_hgs)}"
    if dryrun:
        info(f"Would add {base_msg}.")
    else:
        success(f"Added {base_msg}.")


@app.command(
    "create_hostgroup",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Create a host group with default user group permissions from config",
            "create_hostgroup 'My Host Group'",
        ),
        Example(
            "Create a host group with specific RO and RW user group access",
            "create_hostgroup 'My Host Group' --ro-groups users --rw-groups admins",
        ),
        Example(
            "Create a host group without default user group permissions",
            "create_hostgroup 'My Host Group' --no-usergroup-permissions",
        ),
    ],
)
def create_hostgroup(
    hostgroup: str = typer.Argument(
        help="Name of host group.",
        show_default=False,
    ),
    rw_groups: Optional[str] = typer.Option(
        None,
        "--rw-groups",
        help="User group(s) to give read/write permissions. Comma-separated.",
    ),
    ro_groups: Optional[str] = typer.Option(
        None,
        "--ro-groups",
        help="User group(s) to give read-only permissions. Comma-separated.",
    ),
    use_default_usergroups: bool = typer.Option(
        True,
        "--default-usergroups/--no-default-usergroups",
        "--usergroup-permissions/--no-usergroup-permissions",
        help="Assign default user group permissions from configs.",
    ),
) -> None:
    """Create a new host group.

    Assigns default user group permissions by default.

    * [option]--rw-groups[/] defaults to config option [configopt]app.commands.create_hostgroup.rw_groups[/].
    * [option]--ro-groups[/] defaults to config option [configopt]app.commands.create_hostgroup.ro_groups[/].
    * Use [option]--no-default-usergroups[/] to avoid adding user to default user groups.
    """
    from zabbix_cli.models import Result

    if app.state.client.hostgroup_exists(hostgroup):
        exit_err(f"Host group {hostgroup!r} already exists.")

    with app.status("Creating host group..."):
        hostgroup_id = app.state.client.create_hostgroup(hostgroup)

    app_config = app.state.config.app

    rw_grps: list[str] = parse_list_arg(rw_groups)
    ro_grps: list[str] = parse_list_arg(ro_groups)
    if use_default_usergroups:
        rw_grps.extend(app_config.commands.create_hostgroup.rw_groups)
        ro_grps.extend(app_config.commands.create_hostgroup.ro_groups)

    try:
        # Admin group(s) gets Read/Write
        for usergroup in rw_grps:
            app.state.client.update_usergroup_rights(
                usergroup, [hostgroup], UsergroupPermission.READ_WRITE, hostgroup=True
            )
            info(f"Assigned Read/Write permission for user group {usergroup!r}")
        # Default group(s) gets Read
        for usergroup in ro_grps:
            app.state.client.update_usergroup_rights(
                usergroup, [hostgroup], UsergroupPermission.READ_ONLY, hostgroup=True
            )
            info(f"Assigned Read-only permission for user group {usergroup!r}")
    except Exception as e:
        # All or nothing. Delete group if we fail to assign permissions.
        error(f"Failed to assign permissions to host group {hostgroup!r}: {e}")
        info("Deleting host group...")
        app.state.client.delete_hostgroup(hostgroup_id)
        exit_err(f"Failed to create host group {hostgroup!r}.")

    render_result(Result(message=f"Created host group {hostgroup} ({hostgroup_id})."))


@app.command("extend_hostgroup", rich_help_panel=HELP_PANEL)
def extend_hostgroup(
    src_group: str = typer.Argument(
        help="Group to get hosts from.",
        show_default=False,
    ),
    dest_group: str = typer.Argument(
        help="Group(s) to add hosts to. Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Show hosts and groups without making changes.",
    ),
) -> None:
    """Add all hosts from a host group to other host groups.

    The source group is not modified. Existing hosts in the destination group(s)
    are not removed or modified.
    """
    from zabbix_cli.commands.results.hostgroup import ExtendHostgroupResult

    dest_args = parse_list_arg(dest_group)
    src = app.state.client.get_hostgroup(src_group, select_hosts=True)
    dest = app.state.client.get_hostgroups(*dest_args, select_hosts=True)

    if not dest:
        exit_err(f"No host groups found matching {dest_group!r}.")
    if not src.hosts:
        exit_err(f"No hosts found in host group {src_group!r}.")

    # TODO: calculate the number of hosts that would be added like the other commands
    if not dryrun:
        app.state.client.add_hosts_to_hostgroups(src.hosts, dest)
        success(
            f"Copied {len(src.hosts)} hosts from {src.name!r} to {len(dest)} groups."
        )
    else:
        info(f"Would copy {len(src.hosts)} hosts from {src.name!r}:")
    render_result(ExtendHostgroupResult.from_result(src, dest))


@app.command("move_hosts", rich_help_panel=HELP_PANEL)
def move_hosts(
    src_group: str = typer.Argument(
        help="Group to move hosts from.",
        show_default=False,
    ),
    dest_group: str = typer.Argument(
        help="Group to move hosts to.",
        show_default=False,
    ),
    rollback: bool = typer.Option(
        True,
        "--rollback/--no-rollback",
        help="Rollback changes if hosts cannot be removed from source group afterwards.",
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Show hosts and groups without making changes.",
    ),
) -> None:
    """Move all hosts from one host group to another."""
    from zabbix_cli.commands.results.hostgroup import MoveHostsResult

    src = app.state.client.get_hostgroup(src_group, select_hosts=True)
    dest = app.state.client.get_hostgroup(dest_group, select_hosts=True)

    if not src.hosts:
        exit_err(f"No hosts found in host group {src_group!r}.")

    # TODO: calculate the number of hosts that would be added like the other commands
    if dryrun:
        info(f"Would move {len(src.hosts)} hosts to {dest.name!r}:")
    else:
        app.state.client.add_hosts_to_hostgroups(src.hosts, [dest])
        info(f"Added hosts to {dest.name!r}")
        try:
            app.state.client.remove_hosts_from_hostgroups(src.hosts, [src])
        except Exception as e:
            if rollback:
                error(
                    f"Failed to remove hosts from {src.name!r}. Attempting to roll back changes."
                )
                app.state.client.remove_hosts_from_hostgroups(src.hosts, [dest])
            raise e
        else:
            info(f"Removed hosts from {src.name!r}.")
        success(f"Moved {len(src.hosts)} hosts from {src.name!r} to {dest.name!r}.")

    render_result(MoveHostsResult.from_result(src, dest))


@app.command(
    "remove_host_from_hostgroup",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Remove a host to a host group",
            "remove_host_from_hostgroup 'My host' 'My host group'",
        ),
        Example(
            "Remove multiple hosts from a host group",
            "remove_host_from_hostgroup 'host1,host2' 'My host group'",
        ),
        Example(
            "Remove multiple hosts from multiple host groups",
            "remove_host_from_hostgroup 'host1,host2' 'My host group,Another group'",
        ),
    ],
)
def remove_host_from_hostgroup(
    hostnames_or_ids: str = typer.Argument(
        help="Host names or IDs. Comma-separated. Supports wildcards.",
        metavar="HOSTS",
        show_default=False,
    ),
    hostgroups: str = typer.Argument(
        help="Host group names or IDs. Comma-separated. Supports wildcards.",
        metavar="HOSTGROUPS",
        show_default=False,
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Preview changes",
    ),
) -> None:
    """Remove hosts from host groups."""
    import itertools

    from zabbix_cli.commands.results.hostgroup import RemoveHostsFromHostGroup
    from zabbix_cli.models import AggregateResult

    hosts = parse_hosts_arg(app, hostnames_or_ids)
    hgs = parse_hostgroups_arg(app, hostgroups, select_hosts=True)
    if not dryrun:
        with app.status("Removing hosts from host groups..."):
            app.state.client.remove_hosts_from_hostgroups(hosts, hgs)

    result: list[RemoveHostsFromHostGroup] = []
    for hg in hgs:
        r = RemoveHostsFromHostGroup.from_result(hosts, hg)
        if not r.hosts:
            continue
        result.append(r)

    total_hosts = len(set(itertools.chain.from_iterable((r.hosts) for r in result)))
    total_hgs = len(result)

    render_result(AggregateResult(result=result))
    base_msg = f"{p('host', total_hosts)} from {p('host group', total_hgs)}"
    if dryrun:
        info(f"Would remove {base_msg}.")
    else:
        success(f"Removed {base_msg}.")


@app.command("remove_hostgroup", rich_help_panel=HELP_PANEL)
def delete_hostgroup(
    hostgroup: str = typer.Argument(
        help="Name of host group(s) to delete. Comma-separated.",
        show_default=False,
    ),
    force: bool = typer.Option(
        False,
        "--force",
        help="Remove host group even if it contains hosts.",
    ),
) -> None:
    """Delete a host group."""
    from zabbix_cli.commands.results.hostgroup import HostGroupDeleteResult
    from zabbix_cli.models import Result

    hostgroup_names = parse_list_arg(hostgroup)

    hostgroups = [
        app.state.client.get_hostgroup(hg, select_hosts=True) for hg in hostgroup_names
    ]

    for hg in hostgroups:
        if hg.hosts and not force:
            exit_err(
                f"Host group {hg.name!r} contains {p('host', len(hg.hosts))}. Use --force to delete."
            )
        app.state.client.delete_hostgroup(hg.groupid)

    render_result(
        Result(
            message=f"Host group {hostgroup!r} deleted.",
            result=HostGroupDeleteResult(groups=hostgroup_names),
        ),
    )


@app.command("show_hostgroup", rich_help_panel=HELP_PANEL)
def show_hostgroup(
    ctx: typer.Context,
    hostgroup: str = typer.Argument(
        help="Name of host group.",
        show_default=False,
    ),
) -> None:
    """Show details of a host group."""
    from zabbix_cli.commands.results.hostgroup import HostGroupResult

    hg = app.state.client.get_hostgroup(hostgroup, select_hosts=True)
    render_result(HostGroupResult.from_hostgroup(hg))


# TODO: match V2 behavior
@app.command(
    "show_hostgroups",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Show all host groups (excluding hosts)",
            "show_hostgroups --no-hosts",
        ),
        Example(
            "Show all host groups starting with 'Web-'",
            "show_hostgroups 'Web-*'",
        ),
        Example(
            "Show host groups with 'web' in the name",
            "show_hostgroups '*web*'",
        ),
    ],
)
def show_hostgroups(
    name: str = typer.Argument(
        "*",
        help="Name of host group(s). Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    select_hosts: bool = typer.Option(
        True, "--hosts/--no-hosts", help="Show hosts in each host group."
    ),
    limit: int = OPTION_LIMIT,
) -> None:
    """Show details for host groups.

    Fetching all host groups with hosts can be extremely slow. It is recommended to
    use [option]--no-hosts[/] when not using a specific name filter.
    """
    from zabbix_cli.commands.results.hostgroup import HostGroupResult
    from zabbix_cli.models import AggregateResult

    names = parse_list_arg(name)

    with app.status("Fetching host groups..."):
        hostgroups = app.state.client.get_hostgroups(
            *names,
            select_hosts=select_hosts,
            search=True,
            sort_field="name",
            sort_order="ASC",
            limit=limit,
        )
    render_result(
        AggregateResult(
            result=[HostGroupResult.from_hostgroup(hg) for hg in hostgroups]
        )
    )


@app.command("show_hostgroup_permissions", rich_help_panel=HELP_PANEL)
def show_hostgroup_permissions(
    hostgroups: str = typer.Argument(
        help="Host group name(s). Comma-separated. Supports wildcards.",
        show_default=False,
    ),
) -> None:
    """Show usergroups with permissions for the given hostgroup.

    Shows permissions for all host groups by default.
    """
    from zabbix_cli.commands.results.hostgroup import HostGroupPermissions
    from zabbix_cli.models import AggregateResult

    hg_names = parse_list_arg(hostgroups)

    with app.status("Fetching host groups..."):
        usergroups = app.state.client.get_usergroups()
        hgs = app.state.client.get_hostgroups(
            *hg_names,
            sort_field="name",
            sort_order="ASC",
            select_hosts=False,
            search=True,
        )

    result: list[HostGroupPermissions] = []
    for hg in hgs:
        permissions: list[str] = []
        for usergroup in usergroups:
            if app.api_version >= (6, 2, 0):
                rights = usergroup.hostgroup_rights
            else:
                rights = usergroup.rights
            for right in rights:
                if right.id == hg.groupid:
                    perm = UsergroupPermission.string_from_value(right.permission)
                    permissions.append(f"{usergroup.name} ({perm})")
                    break
        result.append(
            HostGroupPermissions(
                groupid=hg.groupid,
                name=hg.name,
                permissions=permissions,
            )
        )
    return render_result(AggregateResult(result=result))



from __future__ import annotations

from typing import Optional

from zabbix_cli.models import TableRenderable


class AcknowledgeEventResult(TableRenderable):
    """Result type for `acknowledge_event` command."""

    event_ids: list[str] = []
    close: bool = False
    message: Optional[str] = None


class AcknowledgeTriggerLastEventResult(AcknowledgeEventResult):
    """Result type for `acknowledge_trigger_last_event` command."""

    trigger_ids: list[str] = []
